#define UNICODE
#define _UNICODE

#include <windows.h>
#include "guiutil.h"
#include "evidencefile.h"
#include "simplehashr.h"
#include <string.h>
#include <stdio.h>

unsigned char MD5 = 1;
unsigned char SHA1 = 2;
unsigned char SHA256 = 4;

typedef NTSTATUS WINAPI (*RtlGetLastNtStatus_t)();

/* Read the evidence file */
int ReadEvidences(struct SimpleHasher* app, struct HASHLIST** pList)
{
	wchar_t filename[MAX_FILENAME+1];
	int copied = GetWindowText(app->txtFilePath, filename, MAX_FILENAME);
	int retVal = 0;
	HANDLE file = NULL, mapping = NULL;
	char* data = NULL;
		
	if (copied == 0)
	{
		MessageBox(NULL, L"E' necessario scegliere un file contenente le evidenze da cercare.", L"Nessun file di evidenze", MB_ICONEXCLAMATION | MB_OK);
		SetFocus(app->txtFilePath);
		return 1;
	}
	else if (copied >= MAX_FILENAME)
	{
		MessageBox(NULL, L"Il percorso del file delle evidenze e' troppo lungo.\nSpostarlo in un percorso piu' corto.", L"Percorso troppo lungo", MB_ICONEXCLAMATION | MB_OK);
		return 2;
	}
	
	file = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (file == INVALID_HANDLE_VALUE)
	{
		ShowLastError(L"Errore file evidenze", L"Impossibile aprire il file delle evidenze:\n");
		return 3;
	}
	
	/* This shoudl not fail */
	LARGE_INTEGER largeFileSize;
	DWORDLONG fileSize;
	
	if (  ! GetFileSizeEx(file, &largeFileSize))
	{
		ShowLastError(L"Errore file evidenze", L"Impossibile determinare la dimensione del file delle evidenze:\n");
		return 10;	
	}
	fileSize = MAKE_DWORDLONG(largeFileSize.HighPart, largeFileSize.LowPart);
	
	if (fileSize == 0)
	{
		MessageBox(NULL, L"Il file delle evidenze e' vuoto!", L"Errore file delle evidenze", MB_ICONEXCLAMATION | MB_OK);
		return 11;
	}
	
	mapping = CreateFileMapping(file, NULL, PAGE_READONLY, largeFileSize.HighPart, largeFileSize.LowPart, NULL);
	if ( ! mapping )
	{
		ShowLastError(L"Errore file evidenze", L"Impossibile aprire il file delle evidenze:\n");
		retVal = 4;
		goto finalize;
	}
	
	DWORDLONG fileSizeLeft = fileSize, fileSizeDone = 0;
	DWORDLONG bytesDone;
	char* ptr;
	
	/* Step 1: Quick estimation of the lines */
	int lines = 1;
	BOOL lastWasCR = FALSE;
	BOOL lastWasNewLine = TRUE;
	
	while (fileSizeLeft != 0)
	{
		/* How many bytes to map */
		SIZE_T bytesMapped = fileSizeLeft > MAP_MAX ? MAP_MAX : fileSizeLeft;
		
		/* Map the portion of the file */
		DWORD sizeLow = DWORDLONG_LOW(fileSizeDone), sizeHigh = DWORDLONG_HIGH(fileSizeDone);
		data = MapViewOfFile(mapping,  FILE_MAP_READ, sizeHigh, sizeLow, bytesMapped);
			
		/* Handle any error */
		if ( ! data )
		{
			ShowLastError(L"Errore file evidenze", L"Impossibile mappare il file delle evidenze:\n");
			retVal = 5;
			goto finalize;
		}
	
		bytesDone = 0;
		ptr = data;
		while (bytesDone != bytesMapped)
		{
			if ( ( lastWasCR && *ptr != 10) || *ptr == 13 || ( *ptr == 10 && ! lastWasCR ) )
			{
				if ( ! lastWasNewLine)
					lines++;
			}
				
			lastWasCR = *ptr == 13;
			lastWasNewLine = *ptr == 13 || *ptr == 10;
			
			ptr++;
			bytesDone++;
		}
	
		fileSizeLeft -= bytesMapped;
		fileSizeDone += bytesMapped;
		UnmapViewOfFile(data);
	}
	
	if ( lines == 0)
	{
		MessageBox(NULL, L"Il file delle evidenze contiene solo linee vuote!", L"File delle evidenze vuoto", MB_ICONEXCLAMATION | MB_OK);
		retVal = 12;
		goto finalize;
	}
	
	/* Allocate the buffer */
	*pList = (struct HASHLIST *)AllocMem(sizeof(struct HASHLIST) + lines*sizeof(struct HASH));
	
	/* Zero-out the hashes to ease the debug */
	memset(*pList, 0, sizeof(struct HASHLIST) + lines*sizeof(struct HASH));
	
	if ( ! *pList)
	{
		ShowLastError(L"Errore file evidenze", L"Memoria non sufficiente per leggere tutte le evidenze.\n");
		retVal = 6;
		goto finalize;
	}
	
	(*pList)->nHash = 0;
	(*pList)->hashTypes = 0;
	
	/* Step2: Collect the hashes */
	fileSizeLeft = fileSize;
	fileSizeDone = 0;
	
	int line = 1, lineLen = 0, curByte = 0;
	struct HASH* curHash = (*pList)->hashes;
	wchar_t c;
	
	while (fileSizeLeft != 0)
	{
		/* How many bytes to map */
		SIZE_T bytesMapped = fileSizeLeft > MAP_MAX ? MAP_MAX : fileSizeLeft;
		
		/* Map the portion of the file */
		data = MapViewOfFile(mapping, FILE_MAP_COPY | FILE_MAP_READ , DWORDLONG_HIGH(fileSizeDone), DWORDLONG_LOW(fileSizeDone), bytesMapped);
		
		/* Handle any error */
		if ( ! data )
		{
			ShowLastError(L"Errore file evidenze", L"Impossibile mappare il file delle evidenze:\n");
			retVal = 5;
			goto finalize;
		}
		
		ptr = data;
		bytesDone = 0;
		BOOL oneMoreIteration = FALSE;
		
		while (bytesDone != bytesMapped || oneMoreIteration)
		{
			if (oneMoreIteration)
				c = '\n';
			else 
				c = *ptr;
		
			
			/* New lines */
			if (c == 13 || (c == 10 && !lastWasCR) || (c != 10 && lastWasCR))
			{
				/* Check the len of the line */
				if (lineLen != 0 && lineLen != 64 && lineLen != 40 && lineLen != 32 )
				{
					wchar_t message[2048];
					swprintf(message, 1024, L"L'evidenza alla linea %d ha una lunghezza (%d caratteri) non valida (deve essere una tra 32, 40 o 64 caratteri)",line, lineLen);
					MessageBox(NULL, message, L"Errore file evidenze", MB_ICONERROR | MB_OK);
					retVal = 7;
					goto finalize;
				}
				/* Add a new (as-yet empty) HASH item */
				else if (lineLen != 0)
				{
					/* Sanity check, there must be room */
					if (lines == 0)
					{
						MessageBox(NULL, L"Ooops, c'e' un errore in SimpleHashr.\r\nPuoi segnalarcelo a cert-pa@cert-pa.it allegando il file delle evidenze.\r\nGrazie!", L"Bug in SimpleHashr", MB_ICONERROR |MB_OK);
						retVal = 13;
						goto finalize;
					}
					lines--;
					
					if (lineLen == 64)
					{
						(*pList)->hashTypes |= SHA256;
						curHash->hashType = SHA256;
					}
					if (lineLen == 40)
					{
						(*pList)->hashTypes |= SHA1;
						curHash->hashType = SHA1;
					}
					if (lineLen == 32)
					{
						(*pList)->hashTypes |= MD5;
						curHash->hashType = MD5;
					}	
					
					curByte = 0;
					curHash++;
					(*pList)->nHash++;
				}
				
				line++;
				lineLen = 0;
			}
			else if ( ( *ptr == ' ' || *ptr == 10 ) && (lineLen == 0 || lineLen == 32 || lineLen == 40 || lineLen == 64))
			{
				/* Ignore white-space */
			}
			else
			{
				
				lineLen++;
				
				if ( ! ( (c>= '0' && c <= '9') || (c >= 'A' && c <= 'F') || (c >= 'a' && c <= 'f') ) )
				{
					wchar_t message[2048];
					swprintf(message, 1024, L"L'evidenza alla linea %d non e' valida.\nLa colonna %d contiene il carattere non valido '%c'",line, lineLen, c);
					MessageBox(NULL, message, L"Errore file evidenze", MB_ICONERROR | MB_OK);
					retVal = 6;
					goto finalize;
				}
				else 
				{
					int digit = c >= '0' && c <= '9' ? c - '0' : ( c | 0x20 ) - 'a' + 10;
					curHash->hash[curByte] = (curHash->hash[curByte] << 4) | digit;
					
					curByte += 1 - (lineLen & 1);
				}
			}
			
			lastWasCR = c == 13;
		
			if (oneMoreIteration)
				break;
				
		
			ptr++;
			bytesDone++;
			oneMoreIteration = bytesDone == bytesMapped && fileSizeLeft <= MAP_MAX;
		}
	
		fileSizeLeft -= bytesMapped;
		fileSizeDone += bytesMapped;
		UnmapViewOfFile(data);
	}
	
finalize:
		
	if (data)
		UnmapViewOfFile(data);
	if (mapping)
		CloseHandle(mapping);
	if (file)
		CloseHandle(file);	
		
	return retVal;
}
