#define UNICODE
#define _UNICODE

#include "find.h"
#include <string.h>
#include <stdio.h>
#include "guiutil.h"
#include "simplehashr.h"
#include "evidencefile.h"
#include <commctrl.h>
#include "elevate.h"
#include "md5.h"
#include "sha1.h"
#include "sha256.h"

#define MAX_PATH_LEN (10*1024)
typedef BOOL (*FileProc)(WIN32_FIND_DATA* data, wchar_t*path,  LPVOID context, DWORDLONG n);

#define MAX_ERROR 300

void EnumFiles(wchar_t *path, int buff_size, int path_len, FileProc proc, LPVOID context, DWORDLONG* pN)
{
	WIN32_FIND_DATA data;
	wchar_t errorString[MAX_ERROR];
	DWORDLONG n = 0;
	
	if ( ! pN)
		pN = &n;
	
	/* Check there is enough space to add a slash and an asterisk*/
	if (path_len >= buff_size - 2)
		return;
		
	/* If the path terminates with a slash, add an asterisk*/
	if (path_len > 0 && path[path_len - 1] == '\\')
	{
		path[path_len++] = '*';
		path[path_len] = 0;
	}
	else 
	{
		path[path_len++] = '\\';
		path[path_len++] = '*';
		path[path_len] = 0;
	}
	
	/* Start searching */
	HANDLE find = FindFirstFile(path, &data);
	
	/* Remove the askerisk */
	path_len--;
	path[path_len] = 0;
	
	/* Error? */
	if (find == INVALID_HANDLE_VALUE)
	{
		/* Get the last error*/
		wchar_t* msg = L"[WRN] Impossibile controllare il percorso ";
		LastErrorString(errorString, MAX_ERROR);
		DWORD maxErrorBufferLen = wcslen(msg) + MAX_ERROR + path_len;
		
		/* Allocate a buffer to compose the message */
		wchar_t* errorBuffer = AllocString(maxErrorBufferLen);
		
		/* The memory allocation failed? */
		if ( ! errorBuffer)
		{
			/* Fallback to writing to times */
			PostError(msg); 
			PostError(path);
		}
		else 
		{
			/* Compose the message and write it */
			swprintf(errorBuffer, maxErrorBufferLen, L"%s%s: %s", msg, path, errorString);
			PostError(errorBuffer);
			
			FreeMem(errorBuffer);
		}
		
		/* Stop */
		return;
	}

	

	do 
	{	
		/* Pause if asked to (.running is FALSE) */
		while ( ! simpleHasher.running)
		{
			Sleep(300);
		}
	
		/* Be defensive */
		if ( ! data.cFileName)
			continue;
			
		/* Skip this (.) and parent (..) */
		if (wcscmp(data.cFileName, L".") == 0 || wcscmp(data.cFileName, L"..") == 0 )
			continue;
			
		/* Skip junctions */
		if ( (data.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0 )
			continue;	
			
		/* How long is the current file name? Does it fit in the remaining space? */
		int curFilenameLen = wcslen(data.cFileName);
		if (curFilenameLen >= buff_size - path_len )
		{
			/* Show an error if the path has grew too long */
			PostError(L"[WRN] Il seguente percorso e' troppo lungo:");
			PostError(path);
			PostError(data.cFileName);
		}
		else
		{
			/* Add the current filename to the path */
			wcscat(path, data.cFileName);
			
			/* Is it a directory? */
			if (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				/* Recurse */
				EnumFiles(path, buff_size, path_len + curFilenameLen, proc, context, pN);
			}
			else
			{
				/* Call the callback */
				if ( ! proc(&data, path, context, *pN))
					break;
				(*pN)++;
			}
			
			/* Remove the current filename */
			path[path_len] = 0;
		}
		
	}
	while (FindNextFile(find, &data) && ! simpleHasher.stopping);
	
	
	FindClose(find);
	
	CheckHeap();
	
	return;
}

void PostFormattedError(wchar_t* format, wchar_t* fallback, wchar_t* fileName)
{
	wchar_t errorString[MAX_ERROR];
	LastErrorString(errorString, MAX_ERROR);
	
	DWORD allMessageLen = wcslen(errorString) + wcslen(fileName) + wcslen(format);
	wchar_t* allMessage = AllocString(allMessageLen);
	
	if ( ! allMessage)
	{
		/* This screw the file skipped count but it is not meant to be precise */
		PostError(fallback);
		PostError(fileName);
	}
	else
	{
		/* Format the message */
		swprintf(allMessage, allMessageLen, format, fileName, errorString);
		PostError(allMessage);
		FreeMem(allMessage);
	}
}

/* This is a working thread */
DWORD WINAPI CheckHashes(LPVOID lpParameter)
{
	struct SimpleHasher* s = (struct SimpleHasher*)lpParameter;
	struct FQUEUE* q = s->fileQueue;

	int myCount = 0;
	BOOL isEmpty = FALSE;
	void* data;
	HANDLE hMapping = INVALID_HANDLE_VALUE;
	LARGE_INTEGER liFileSize;
	DWORDLONG n = 0;
	
	wchar_t fileName[MAX_FILENAME];
	wchar_t perc[MAX_ERROR];
	
	MD5_CTX md5Ctx;
	SHA1_CTX sha1Ctx;
	SHA256_CTX sha256Ctx;
	
	struct HASH md5Hash = { .hashType = MD5, .hash = {0xd4, 0x1d, 0x8c, 0xd9, 0x8f, 0x00, 0xb2, 0x04, 0xe9, 0x80, 0x09, 0x98, 0xec, 0xf8, 0x42, 0x7e} };
	struct HASH sha1Hash = { .hashType = SHA1, .hash = {0xda, 0x39, 0xa3, 0xee, 0x5e, 0x6b, 0x4b, 0x0d, 0x32, 0x55, 0xbf, 0xef, 0x95, 0x60, 0x18, 0x90, 0xaf, 0xd8, 0x07, 0x09} };
	struct HASH sha256Hash = { .hashType = SHA256, .hash = {0xe3, 0xb0, 0xc4, 0x42, 0x98, 0xfc, 0x1c, 0x14, 0x9a, 0xfb, 0xf4, 0xc8, 0x99, 0x6f, 0xb9, 0x24, 0x27, 0xae, 0x41, 0xe4, 0x64, 0x9b, 0x93, 0x4c, 0xa4, 0x95, 0x99, 0x1b, 0x78, 0x52, 0xb8, 0x55 } };
	
	do
	{	
		/* Get the next handle */
		HANDLE hFile = INVALID_HANDLE_VALUE;
		while ( ! GetItem(q, fileName, &hFile))
			SwitchToThread();
		
		/* INVALID_HANDLE_VALUE means we must exit */
		if (hFile == INVALID_HANDLE_VALUE)
			break;
		
		n++;
			
		/* Get the file size */
		if ( ! GetFileSizeEx(hFile, &liFileSize))
		{
			PostFormattedError(L"[WRN] Impossibile determinare la dimensione del file %s: %s", L"Impossibile determinare la dimensione del file:", fileName);
			CloseHandle(hFile);
			continue;
			
		}
		
		/* Is the file empty ? */
		isEmpty = liFileSize.LowPart == 0 && liFileSize.HighPart == 0;
		
		if ( ! isEmpty)
		{
			hMapping = INVALID_HANDLE_VALUE;
			data = NULL;
			
			/* Create the file mapping */
			hMapping = CreateFileMapping(hFile, NULL, PAGE_READONLY, liFileSize.HighPart, liFileSize.LowPart, NULL);
					
			if ( ! hMapping || hMapping == INVALID_HANDLE_VALUE)
			{
				PostFormattedError(L"[WRN] Impossibile leggere il file %s: %s", L"Impossibile leggere il file:", fileName);

				CloseHandle(hFile);
				continue;
			}
			
			/* Map the file and compute the hashes */
			DWORDLONG fileSizeLeft = MAKE_DWORDLONG(liFileSize.HighPart, liFileSize.LowPart), fileSizeDone = 0;
			BOOL valid = TRUE;
		
			//(void)md5Ctx; (void)sha1Ctx; (void)sha256Ctx;
		
			/* Init the hashes */
			if (s->hashList->hashTypes & MD5)
				MD5_Init(&md5Ctx);
				
			if (s->hashList->hashTypes & SHA1)
				SHA1Init(&sha1Ctx);
			
			if (s->hashList->hashTypes & SHA256)
				sha256_init(&sha256Ctx);
				
			while (fileSizeLeft != 0)
			{
				/* How many bytes to map */
				SIZE_T bytesMapped = fileSizeLeft > MAP_MAX ? MAP_MAX : fileSizeLeft;
				
				/* Map the portion of the file */
				DWORD sizeLow = DWORDLONG_LOW(fileSizeDone), sizeHigh = DWORDLONG_HIGH(fileSizeDone);
				data = MapViewOfFile(hMapping,  FILE_MAP_READ, sizeHigh, sizeLow, bytesMapped);
					
				/* Handle any error */
				if ( ! data )
				{
					PostFormattedError(L"[WRN] Impossibile mappare il file %s: %s", L"Impossibile mappare il file:", fileName);
					valid = FALSE;
					break;
				}
			
				if (s->hashList->hashTypes & MD5)
					MD5_Update(&md5Ctx, data, bytesMapped);	
					
				if (s->hashList->hashTypes & SHA1)
					SHA1Update(&sha1Ctx, data, bytesMapped);
						
				if (s->hashList->hashTypes & SHA256)
					sha256_update(&sha256Ctx, data, bytesMapped);
								
				fileSizeLeft -= bytesMapped;
				fileSizeDone += bytesMapped;
				UnmapViewOfFile(data);
				
				/* Report */
				if (s->totalSize != 0)
				{
					EnterCriticalSection(&s->critical_section);
						s->curSize += bytesMapped;
					
						if (n % 20 == 1)
						{
							ToPerc(perc, MAX_ERROR, &s->totalSize, &s->curSize, s->skippedFiles);
							PostStatus(perc);
						}
					LeaveCriticalSection(&s->critical_section);
				}
				
			}
			
			CloseHandle(hMapping);
			
			if ( ! valid)
			{
				CloseHandle(hFile);
				continue;
			}
			
			if (s->hashList->hashTypes & MD5)
				MD5_Final(md5Hash.hash, &md5Ctx);
				
			if (s->hashList->hashTypes & SHA1)
				SHA1Final(sha1Hash.hash, &sha1Ctx);
				
			if (s->hashList->hashTypes & SHA256)
				sha256_final(&sha256Ctx, sha256Hash.hash);

		}
		else
		{
			/* Empty file */
		}
		
		CloseHandle(hFile);
	
		/* Show the matches */
		wchar_t* resultFormat = L"[***] EVIDENZA %s TROVATA IN %s";
		DWORD resultLen = wcslen(fileName) + 64 + wcslen(resultFormat);
		wchar_t* resultBuffer = NULL;
		wchar_t hashString[65] = {0};
		
		if (s->hashList->hashTypes & MD5)
			if (bsearch(&md5Hash, s->hashList->hashes, s->hashList->nHash, sizeof(struct HASH), &cmpHash))
			{
				resultBuffer = AllocString(resultLen);
				if ( ! resultBuffer)
				{
					PostLog(hashString);
					PostLog(fileName);
				}
				else
				{
					HashToString(hashString, md5Hash.hash, 16);
					swprintf(resultBuffer, resultLen, resultFormat, hashString, fileName);
					PostLog(resultBuffer);
					FreeMem(resultBuffer);
				}
				
			}
			
			
		if (s->hashList->hashTypes & SHA1)
			if (bsearch(&sha1Hash, s->hashList->hashes, s->hashList->nHash, sizeof(struct HASH), &cmpHash))
			{
				resultBuffer = AllocString(resultLen);
				if ( ! resultBuffer)
				{
					PostLog(hashString);
					PostLog(fileName);
				}
				else
				{
					HashToString(hashString, sha1Hash.hash, 20);
					swprintf(resultBuffer, resultLen, resultFormat, hashString, fileName);
					PostLog(resultBuffer);
					FreeMem(resultBuffer);
				}
			}
			
		if (s->hashList->hashTypes & SHA256)	
			if (bsearch(&sha256Hash, s->hashList->hashes, s->hashList->nHash, sizeof(struct HASH), &cmpHash))
			{
				resultBuffer = AllocString(resultLen);
				if ( ! resultBuffer)
				{
					PostLog(hashString);
					PostLog(fileName);
				}
				else
				{
					HashToString(hashString, sha256Hash.hash, 32);
					swprintf(resultBuffer, resultLen, resultFormat, hashString, fileName);
					PostLog(resultBuffer);
					FreeMem(resultBuffer);
				}
			}	
			
		/* Report fileName if no percentage available */
		if (s->totalSize == 0 && n % 50 == 1)
			PostStatus(fileName);
		
	}
	while (TRUE);
	
	
	/*No more file to check, decrement the working thread count */
	EnterCriticalSection(&s->critical_section);
		(s->nHashingThreads)--;
		myCount = s->nHashingThreads;
	LeaveCriticalSection(&s->critical_section);
	
	/* The last thread set the event */
	if (myCount == 0)
	{
		SetEvent(s->eventDone);
	}
	
	ExitThread(0);
	return 0;
}


/* Callback for counting the total file size on a path */
BOOL GetTotalSize(WIN32_FIND_DATA* data, wchar_t* path, LPVOID context, DWORDLONG n)
{
	(void)path;
	
	DWORDLONG* totalSize = (DWORDLONG*)context;
	
	*totalSize += MAKE_DWORDLONG(data->nFileSizeHigh, data->nFileSizeLow);
	
	if (n % 100 == 0)
	{
		EnterCriticalSection(&simpleHasher.critical_section);
			PostStatus(path);
		LeaveCriticalSection(&simpleHasher.critical_section);
	}
	
	return TRUE;
}

/* Callback for adding the file handle to the queue */
BOOL FindAllFiles(WIN32_FIND_DATA* data, wchar_t* path, LPVOID context, DWORDLONG n)
{
	(void)context; (void)n;
	
	wchar_t errorString[MAX_ERROR];
	wchar_t* format = L"[WRN] Impossibile accedere al file %s: %s";
	
	HANDLE hFile = CreateFile(path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		LastErrorString(errorString, MAX_ERROR);
		DWORD totalLen = wcslen(errorString) + wcslen(path) + wcslen(format);
		wchar_t* allMessage = AllocString(totalLen);
		
		if ( ! allMessage)
		{
			PostError(L"[WRN] Impossibile accedere al file");
			PostError(path);
		}
		else
		{
			swprintf(allMessage, totalLen, format, path, errorString);
			PostError(allMessage);
			FreeMem(allMessage);
		}
		
	
		/* Add the file size to the count */
		EnterCriticalSection(&simpleHasher.critical_section);
			simpleHasher.curSize += MAKE_DWORDLONG(data->nFileSizeHigh, data->nFileSizeLow);
		LeaveCriticalSection(&simpleHasher.critical_section);
	
	}
	
	else
	{
		/* Add the file to the queue */
		while ( ! AddItem(simpleHasher.fileQueue, hFile, path))
			SwitchToThread();
	}
	return TRUE;
}

/* Thread started for each volume */
DWORD WINAPI CollectVolumeSize(LPVOID lpParameter)
{
	struct COLLECTSIZE* info = (struct COLLECTSIZE*)lpParameter;
	DWORDLONG volumeSize = 0;
	
	/* Privileges will be droppes as soon as this thread exits */
	if 	( isElevated())
		ImpersonateSystem();

	/* Make space for the current path (plus two chars) */
	wchar_t* path = AllocString(MAX_PATH_LEN + 2);
	
	/* The first path is the root*/
	wcscpy(path, info->root);
	
	/* Get the total size in .totalSize */
	EnumFiles(path, MAX_PATH_LEN, wcslen(info->root), &GetTotalSize, &volumeSize, NULL);
	
	/* Done, if we were not stopped, show something to the user */
	if ( ! simpleHasher.stopping)
	{
		/* Show we have done with this volume */
		wchar_t stringSize[MAX_ERROR];
		SizeToStr(stringSize, MAX_ERROR, volumeSize);
		swprintf(path, MAX_PATH_LEN,  L"%s contiene %s di dati", info->root, stringSize);
		PostLog(path);
	}
	
	/* Decrement the thread count */
	EnterCriticalSection(info->critical_sec);
		--*(info->count);
		int myCount = *(info->count);
		
		simpleHasher.totalSize += volumeSize;
	LeaveCriticalSection(info->critical_sec);
	
	/* Free the buffers (we own lpParameter) */
	FreeMem(path);
	FreeMem(info);
	
	/* If this is the last thread that was collecting information, signal the main thread to start the find */
	if (myCount == 0)
	{
		PostMessage(simpleHasher.wndMain, simpleHasher.stopping ? UM_STOPPED : UM_FIND, 0, 0);
	}
	
	ExitThread(0);
	return 0;
}

/* Thread started for each volume when checking the evidences */
DWORD WINAPI CheckVolumeFiles(LPVOID lpParameter)
{
	struct COLLECTSIZE* info = (struct COLLECTSIZE*)lpParameter;
	
	/* Privileges will be droppes as soon as this thread exits */
	if 	( isElevated())
		ImpersonateSystem();

	/* Make space for the current path */
	wchar_t* path = AllocMem(MAX_PATH_LEN);
	
	/* The first path is the root */
	wcscpy(path, info->root);
	
	/* Check all the files */
	EnumFiles(path,  MAX_PATH_LEN - 1, wcslen(info->root), FindAllFiles, NULL, NULL);
	
	/* Add decrement the count */
	EnterCriticalSection(info->critical_sec);
	
		--*(info->count);
		int myCount = *(info->count);
		
		swprintf(path, MAX_PATH_LEN, L"Ricerca in %s finita (ancora %d drive in esecuzione)", info->root, myCount);
		/* AddToLog(path); */
	LeaveCriticalSection(info->critical_sec);
	
	/* Free the buffers (we own lpParameter) */
	
	FreeMem(info);
	
	
	/* If this is the last thread that was collecting files, stop the queue and wait for the working threads to exit */
	if (myCount == 0)
	{
		swprintf(path, MAX_PATH_LEN, L"Ricerca finita, fermo i thread di lavoro");
		/* PostLog(path); */
		
		/* Stop the queue, this will stop each thread */
		StopQueue(simpleHasher.fileQueue, simpleHasher.stopping);
		
		/* Wait for each thread to exit */
		WaitForSingleObject(simpleHasher.eventDone, INFINITE);
		
		swprintf(path, MAX_PATH_LEN, L"Thread di lavoro fermati");
		/* PostLog(path); */
		
		/* Time elapsed */
		DWORD endTime = GetTickCount();
		if (simpleHasher.startTime > endTime)
			simpleHasher.startTime = endTime + (~simpleHasher.startTime + 1);
		else
			simpleHasher.startTime = endTime - simpleHasher.startTime;
		
		/* Notify the UI */
		PostMessage(simpleHasher.wndMain, simpleHasher.stopping ? UM_STOPPED : UM_DONE, 0, 0);
	}
	
	FreeMem(path);
	ExitThread(0);
	return 0;
}


/* Find the evidences */

void FindEvidences(struct SimpleHasher* app)
{
	app->totalSize = 0;
	app->skippedFiles = 0;
	app->hLogFile = OpenLogFile();
	
	/* Log no. and type of evidences */
	wchar_t tmp[MAX_ERROR];
	BOOL hasMD5 = app->hashList->hashTypes & MD5, hasSHA1 = app->hashList->hashTypes & SHA1, hasSHA256 = app->hashList->hashTypes & SHA256;
	
	swprintf(tmp, MAX_ERROR, L"Numero di evidenze da cercare: %d\r\nTipo di evidenze: ", app->hashList->nHash);
	
	if (hasMD5)
		wcscat(tmp, L"MD5");
	
	if (hasSHA1)
		wcscat(tmp,  hasMD5 ? L", SHA1" : L" SHA1");
	
	if (hasSHA256)
		wcscat(tmp,  (hasMD5 || hasSHA1) ? L", SHA256" : L" SHA256");
		
	AddToLog(tmp);
	
	/* If the user want a progress, we need to know how many bytes are there on the disk */
	if (ShowPercent())
	{
		AddToLog(L"Raccolgo i dati necessari per il calcolo della percentuale...");
		
		/* Start a thread for each selected volume */
		InitializeCriticalSection(&app->critical_section);
	
		TVITEM tv = {0};
	    tv.mask = TVIF_TEXT | TVIF_HANDLE;
	    
		HTREEITEM node = TreeView_GetRoot(app->tvwDrives);
		for (; node; node = TreeView_GetNextSibling(app->tvwDrives, node))
		{
			if ( ! TreeView_GetCheckState(app->tvwDrives, node))
				continue;
				
			struct COLLECTSIZE* info = AllocMem(sizeof(struct COLLECTSIZE));
			info->count = &app->count;
			info->critical_sec = &app->critical_section;
			info->total = &app->totalSize;
			
			tv.hItem = node;
	     	tv.pszText = info->root;
     		tv.cchTextMax = MAX_PATH;	
			TreeView_GetItem(app->tvwDrives, &tv);
			
			CreateThread(NULL, 0, &CollectVolumeSize, info, 0, NULL);
		}
		
		
	}
	/* Otherwise just start the search */
	else 
	{
		PostMessage(simpleHasher.wndMain, UM_FIND, 0, 0);
	}
}


void DoFindEvidences(struct SimpleHasher* app)
{
	app->curSize = 0;
	app->count = 0;
	app->encryptedEvidence = EncryptedEvidences();
	
	/* Start and pin a thread for each logical CPU (Don't need to care about threads vs cores) */
	SYSTEM_INFO sinfo;
	GetSystemInfo(&sinfo);
	
	/* Up to 32 CPUs, we don't handle process group yet */
	for (unsigned int i = 0; i < sinfo.dwNumberOfProcessors && i < 32; i++)
	{
		HANDLE hThread = CreateThread(NULL, 0, CheckHashes, app, CREATE_SUSPENDED, NULL);
		app->nHashingThreads++;
		SetThreadAffinityMask(hThread, (1 << (i % 32)) );
		ResumeThread(hThread);
	}
	
	wchar_t threadInfo[MAX_ERROR];
	swprintf(threadInfo, MAX_ERROR, L"%d Thread di lavoro partiti", app->nHashingThreads);
	/* PostLog(threadInfo); */
	
	/* Start a thread for each selected volume */
	

	TVITEM tv = {0};
    tv.mask = TVIF_TEXT | TVIF_HANDLE;
    
    /* Count first */
    HTREEITEM node = TreeView_GetRoot(app->tvwDrives);
	for (; node; node = TreeView_GetNextSibling(app->tvwDrives, node))
	{
		if ( ! TreeView_GetCheckState(app->tvwDrives, node))
			continue;
			
		(app->count)++;
	}
	
	int threadCount = app->count;
	
	node = TreeView_GetRoot(app->tvwDrives);
	for (; node; node = TreeView_GetNextSibling(app->tvwDrives, node))
	{
		if ( ! TreeView_GetCheckState(app->tvwDrives, node))
			continue;
			
		struct COLLECTSIZE* info = AllocMem(sizeof(struct COLLECTSIZE));
		info->count = &app->count;
		info->critical_sec = &app->critical_section;
		info->total = &app->curSize;
		
		tv.hItem = node;
     	tv.pszText = info->root;
 		tv.cchTextMax = MAX_PATH;	
		TreeView_GetItem(app->tvwDrives, &tv);
		
		CreateThread(NULL, 0, &CheckVolumeFiles, info, 0, NULL);
	}
	
	swprintf(threadInfo, MAX_ERROR, L"%d Thread di ricerca partiti", threadCount);
	/* PostLog(threadInfo); */
	
	app->startTime = GetTickCount();
}









