#include "guiutil.h"
#include "messageloop.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define CHAR_WIDTH 2

#define DEBUG 1

BOOL CheckHeap()
{
	BOOL res =  HeapValidate(GetProcessHeap(), 0, NULL) != 0;
	/*if ( ! res)
		MessageBox(NULL, L"Heap is invalid!", L"OMG", 0x10);*/
	return res;
}

void* AllocMem(SIZE_T size)
{
	//void* tmp =  malloc(size);
	void* tmp = HeapAlloc(GetProcessHeap(), 0, size );
	
	if ( ! tmp && DEBUG)
	{
		MessageBox(NULL, L"Memoria esaurita!", L"Errore", MB_ICONERROR | MB_OK);
	}
	
	return tmp;
}

/* Allocate space for size many chars plus the null term. Independently of UNICODE */
void* AllocString(SIZE_T size)
{	
	return AllocMem((size + 1) * CHAR_WIDTH);
}

void FreeMem(void* base)
{
	//free(base); 
	HeapFree(GetProcessHeap(), 0, base);
}

/* If show is TRUE, it hides everything but the log edit control and the start/pause buttons */
void ShowLog(BOOL show)
{
	UINT sLog = show ? SW_SHOW : SW_HIDE, sOther = show ? SW_HIDE : SW_SHOW;
	
	ShowWindow(simpleHasher.txtLog, sLog);
	ShowWindow(simpleHasher.btnBrowse, sOther);
	ShowWindow(simpleHasher.txtFilePath, sOther);
	/*ShowWindow(simpleHasher.chkEncrypted, sOther);*/
	ShowWindow(simpleHasher.chkShowPercent, sOther);
	ShowWindow(simpleHasher.lblSelectDrive, sOther);
	ShowWindow(simpleHasher.lblSelectFile, sOther);
	ShowWindow(simpleHasher.tvwDrives, sOther);
	ShowWindow(simpleHasher.btnBrowse, sOther);
	ShowWindow(simpleHasher.lblProgress, sLog);
	
	SetWindowText(simpleHasher.txtLog, L"");
	SetWindowText(simpleHasher.btnStart, L"Inizia");
}

/* Tell if the user requested for encrypted evidences */
BOOL EncryptedEvidences()
{
	return SendMessage(simpleHasher.chkEncrypted, BM_GETCHECK, 0, 0) == BST_CHECKED;
}

/* Tell if the user asked for the percentage */
BOOL ShowPercent()
{
	return SendMessage(simpleHasher.chkShowPercent, BM_GETCHECK, 0, 0) == BST_CHECKED;
}

/* Add to the log control*/
void AddToLog(wchar_t* str)
{
	SendMessage(simpleHasher.txtLog, EM_SETSEL, 0, -1);
	SendMessage(simpleHasher.txtLog, EM_SETSEL, -1, -1);
	SendMessage(simpleHasher.txtLog, EM_REPLACESEL, FALSE, (LPARAM)str);
	SendMessage(simpleHasher.txtLog, EM_SETSEL, -1, -1);
	SendMessage(simpleHasher.txtLog, EM_REPLACESEL, FALSE, (LPARAM)L"\r\n");
	SendMessage(simpleHasher.txtLog, EM_SETSEL, -1, -1);
	SendMessage(simpleHasher.txtLog, EM_SCROLLCARET, 0, 0);
}

/* Open the log file in the same path as the executable */
HANDLE OpenLogFile()
{
	wchar_t* logFileName = L"simpleHasher.log";
	wchar_t fileName[4096];
	DWORD exePathLen = GetModuleFileName(NULL, fileName, 4096);
	
	/* Get the filename */
	if ( exePathLen == 0 || GetLastError() == ERROR_INSUFFICIENT_BUFFER)
		return INVALID_HANDLE_VALUE;
	
	/* Find the last slash and append the log file name */
	for (; exePathLen != 0xffffffff; exePathLen--)
		if (fileName[exePathLen] == '\\')
			break;
	fileName[exePathLen+1] = 0;
	wcscat(fileName, logFileName);
	
	/* Truncate the existing log */
	HANDLE hFile = CreateFile(fileName, GENERIC_WRITE,  FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL,  NULL);
		
	return hFile;
}

/* Write to the file */
void PostError(wchar_t* str)
{
	void* data;
	DWORD dataLen, lenWritten;
	char* mbString = NULL;
	
	/* Exit early if the file is not open */
	if (simpleHasher.hLogFile == INVALID_HANDLE_VALUE)
		return;
	
	/* How many bytes do we need? */	
	DWORD requiredLen = WideCharToMultiByte(CP_UTF8, 0, str, -1, NULL, 0, NULL, NULL);
	
	/* If it failed, fallback to writing the Unicode string (better than nothing) */
	if (requiredLen == 0)
	{
		data = str;
		dataLen = wcslen(str) * 2;
	}
	else 
	{
		/* Convert and string and fallback if needed */
		mbString = AllocMem(requiredLen + 1);
		if (WideCharToMultiByte(CP_UTF8, 0, str, -1, mbString, requiredLen, NULL, NULL) == 0)
		{
			data = str;
			dataLen = wcslen(str) * 2;
		}
		else
		{
			data = mbString;
			dataLen = requiredLen - 1;
		}
	}
	
	/* Write to the log file */
	EnterCriticalSection(&simpleHasher.critical_section);
		if (simpleHasher.skippedFiles > 0)
			WriteFile(simpleHasher.hLogFile, "\r\n", 2, &lenWritten, NULL);
		WriteFile(simpleHasher.hLogFile, data, dataLen, &lenWritten, NULL);
		
		/* Increment the number of skipped files */
		simpleHasher.skippedFiles++;
	LeaveCriticalSection(&simpleHasher.critical_section);
	
	/* Free */
	if (mbString)
		FreeMem(mbString);	
}

/* Add a string to the log control (Synchronized) */
void PostLog(wchar_t* str)
{
	EnterCriticalSection(&simpleHasher.critical_section);
	AddToLog(str);
	LeaveCriticalSection(&simpleHasher.critical_section);
}

/* Set a string in the status control (Synchronized) */
void PostStatus(wchar_t* str)
{
	EnterCriticalSection(&simpleHasher.critical_section);
	SetWindowText(simpleHasher.lblProgress, str);
	LeaveCriticalSection(&simpleHasher.critical_section);

}



/* Convert a 64-bit quantity into a size string */
void SizeToStr(wchar_t* str, DWORD len, DWORDLONG size)
{
	if (size < 1024)
		swprintf(str, len, L"%lu B", size);
	else if (size < (1<<20) )
		swprintf(str, len, L"%.3f KiB", (double)size / (1 << 10));
	else if (size < (1<<30) )
		swprintf(str, len, L"%.3f MiB", (double)size /  (1 << 20));
	else if (size < ((DWORDLONG)1<<40) )
		swprintf(str, len, L"%.3f GiB", (double)size /  (1 << 30));
	else
		swprintf(str, len, L"%.3f TiB", (double)size /  ((DWORDLONG)1 << 40));
}

/* Make the percent ratio between two 64-bit quantities */
void ToPerc(wchar_t* str, DWORD size, DWORDLONG* all, DWORDLONG* cur, int skipped)
{
	DWORDLONG theAll = *all >> 20;
	DWORDLONG theCur = *cur >> 20;
	
	swprintf(str, size, L"Ricerca in corso: %.2f%% (%d file sono stati saltati)", (double)theCur / theAll * 100, skipped);
}

/* Format the last error as a string */
void LastErrorString(wchar_t* errorString, DWORD size)
{
	DWORD lastErr = GetLastError();
	DWORD len = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ARGUMENT_ARRAY, NULL, lastErr, 0, errorString, size - 1, NULL);
	if (len > 0)
	{
		errorString[len-1] = errorString[len-2] = 0;
	}
	else
		swprintf(errorString, size - 1, L"0x%08x", lastErr);
}

/* Show the last error in a message box */
void ShowLastError(wchar_t* title, wchar_t* message)
{
	/* Error as a string */
	wchar_t errorString[1024];
	LastErrorString(errorString, 1024);
	
	DWORD totalLen = wcslen(errorString) + wcslen(message);
	wchar_t* allMessage = AllocString(totalLen);
	
	if ( ! allMessage)
	{
		MessageBox(NULL, message, title, MB_ICONERROR | MB_OK);
	}
	else 
	{
		wcscpy(allMessage, message);
		wcscat(allMessage, errorString);
		MessageBox(NULL, allMessage, title, MB_ICONERROR | MB_OK);
		FreeMem(allMessage);
	}
}

/* Enumerate the volumes */
int EnumVolumes(struct SimpleHasher* simpleHasher)
{
	/* Get the logical drives */
	DWORD requiredLen = GetLogicalDriveStrings(0, NULL);
	if ( ! requiredLen)
		return 1;
		
	LPWSTR buffer = AllocString(requiredLen);
	void* originalBuffer = buffer;
	if ( ! buffer)
		return 2;
	
	if ( ! GetLogicalDriveStrings(requiredLen, buffer))
	{
		FreeMem(buffer);
		return 3;
	}
		
	/* Scan the drives */	
	while (*buffer)
	{
		TV_INSERTSTRUCT insert;
		insert.hParent = (HTREEITEM)TVI_ROOT; 
		insert.hInsertAfter = (HTREEITEM)TVI_SORT; 
				
		insert.item.mask = TVIF_TEXT | TVIF_STATE;
		insert.item.state = INDEXTOSTATEIMAGEMASK(2);
		insert.item.stateMask = TVIS_STATEIMAGEMASK;
		insert.item.pszText = buffer; 
		insert.item.cchTextMax = 0; 
				
		/* Insert the item */
		HTREEITEM hItem = (HTREEITEM)SendMessage(simpleHasher->tvwDrives,  TVM_INSERTITEM, 0, (LPARAM)&insert);
		
		/* Check the item */
		TVITEM item;
		item.hItem = hItem;
		item.mask = TVIF_STATE | TVIF_HANDLE;
		item.stateMask = TVIS_STATEIMAGEMASK;
		item.state = INDEXTOSTATEIMAGEMASK(2);

		SendMessage(simpleHasher->tvwDrives, TVM_SETITEM, 0,  (LPARAM)&item);
		
		/* Go to the next drive */
		while (*buffer++);
	}
	
	FreeMem(originalBuffer);
	
	return 0;
	
}

int cmpHash(const void *a, const void *b)
{
	struct HASH* h1 = (struct HASH*)a, *h2 = (struct HASH*)b;
	int h1Len = h1->hashType == MD5 ? 16 : (h1->hashType == SHA1 ? 20 : 32);
	int h2Len = h2->hashType == MD5 ? 16 : (h2->hashType == SHA1 ? 20 : 32);
	
	if (h1->hashType != h2->hashType)
		return h1->hashType-h2->hashType;
	
	for (int i = 0; i < h1Len; i++)
	{
		if (i >= h2Len || h1->hash[i] > h2->hash[i])
			return 1;
		
		if (h1->hash[i] < h2->hash[i])
			return -1;
	}
	
	return h2Len > h1Len ? -1 : 0;
}

void HashToString(wchar_t* dest, unsigned char* hash, int len)
{
	wchar_t digits[] = L"0123456789abcdef";
	for (int i = 0; i < len * 2; i++)
		*dest++ = digits[ ( hash[i >> 1] >> 4*( (i + 1) & 1) ) & 0x0f ];
	*dest = 0;
}



