#define _WIN32_IE 0x0301

#define _UNICODE
#define UNICODE

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <string.h>

#include "messageloop.h"
#include "guiutil.h"
#include "simplehashr.h"
#include "evidencefile.h"
#include "elevate.h"
#include "queue.h"

/* Represent the state of the application */
struct SimpleHasher simpleHasher = { .running = FALSE, .timer = 0, .started = FALSE, .hashList = NULL, .totalSize = 0, .done = FALSE };

/* ENTRY POINT */

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd )
{
	(void)hPrevInstance; (void)lpCmdLine; (void)nShowCmd;
	
	/* Check this is the only instance */
	HANDLE mutex = CreateMutex(NULL, TRUE, L"SimpleHashr");
	if (GetLastError() == ERROR_ALREADY_EXISTS)
	{
		MessageBox(NULL, L"Un'altra instanza di SimpleHashr e' in esecuzione.", L"SimpleHashr", MB_ICONERROR | MB_OK);
		return 1;
	}
		
	/* State init */
	InitializeCriticalSection(&simpleHasher.critical_section);	
	InitFQueue(&simpleHasher.fileQueue, 128);	
	simpleHasher.eventDone = CreateEvent(NULL, FALSE, FALSE, NULL);
	
	/* Visual styles */
	INITCOMMONCONTROLSEX icce = { sizeof(icce), ICC_STANDARD_CLASSES };
	InitCommonControlsEx(&icce); 	
	
		
	/* Create the main window*/	
	int error;
	if ( (error = CreateMainWindow(hInstance, &simpleHasher, MainWinTable)) != 0 )	
	{
		MessageBox(NULL, L"Impossibile creare la finestra principale di SimpleHashr.", L"SimpleHashr", MB_ICONERROR | MB_OK);
		return error;
	}
	
	/*Dispatch the messages*/
	MessageLoop(hInstance);	
		
	DeleteCriticalSection(&simpleHasher.critical_section);
	FreeFQueue(simpleHasher.fileQueue);
	CloseHandle(simpleHasher.eventDone);
	
	/* Not necessary but be nice*/
	ReleaseMutex(mutex);
	CloseHandle(mutex);
	
	return 0;
}




