#include <windows.h>
#include "messageloop.h"
#include "events.h"
#include "simplehashr.h"
#include "monitors.h"

/* Maps the messages to the events */
struct MSG_HANDLER MainWinTable[] = {
	
	{ WM_CLOSE, &Application_OnClose },
	{ WM_SIZE, &Application_OnResize },
	{ WM_SHOWWINDOW, &Application_OnShow },
	{ WM_GETMINMAXINFO, &Application_OnMinMax },
	{ WM_COMMAND, &Application_OnClick },
	{ WM_TIMER, &Application_OnTimer },
	{ WM_DROPFILES, &Application_OnDropFiles },
	{ UM_STOPPED, &Application_OnStopped },
	{ UM_FIND, & Application_OnDoFind },
	{ UM_DONE, & Application_OnDone },
	
	{ 0, (WNDPROC)NULL }
};



/* Register a class and create a window */
int CreateMainWindow(HINSTANCE hInstance, struct SimpleHasher* simpleHasher, struct MSG_HANDLER* msgTable)
{
	static LPSTR className = "SimpleHashrClass";
	
	WNDCLASSEX class = {
		sizeof(class), 
		CS_DBLCLKS | CS_DROPSHADOW | CS_GLOBALCLASS | CS_OWNDC, //| CS_VREDRAW | CS_HREDRAW,
		TableWinProc,
		0, sizeof(msgTable),
		hInstance, 
		LoadIcon(hInstance, MAKEINTRESOURCE(2)), LoadCursor(hInstance, IDC_ARROW),
		(HBRUSH)(COLOR_BTNFACE + 1),
		NULL,
		className, 
		NULL };
		
	if ( ! RegisterClassEx(&class))
		return 1;
	
	/* Put the window on the primary display (where else? this is the first window)*/
	RECT primary = GetPrimaryMonitor();	
	int winWidth = (primary.right-primary.left) * 6 / 10, winHeight = (primary.bottom-primary.top) * 7 / 10;
	int winX = (primary.right-primary.left) * 2 / 10, winY = (primary.bottom-primary.top) * 15 / 100;
	
	simpleHasher->wndMain = CreateWindowEx(
		WS_EX_ACCEPTFILES | WS_EX_APPWINDOW, className, "SimpleHashr", 
		WS_TILEDWINDOW | WS_CLIPCHILDREN,
		winX, winY, winWidth, winHeight, NULL, NULL, hInstance, NULL);
		
	if ( ! simpleHasher->wndMain)
		return 2;
		
	/* Save the message table */		
	SetWindowLongPtr(simpleHasher->wndMain, 0, (LONG_PTR)msgTable);
		
		
	/* Make a font */
	HFONT hFont = CreateFont(24, 0, 0, 0, FW_EXTRALIGHT, FALSE, FALSE, FALSE, 
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,CLEARTYPE_QUALITY, (FF_ROMAN << 28) | DEFAULT_PITCH, "Times New Roman" );
		
	/* Create the components */	
	
	/*Label*/
	simpleHasher->lblSelectDrive = CreateWindowEx(0, "STATIC", "Seleziona le unit disco da analizzare", WS_CHILD | WS_VISIBLE, 0, 0, 0, 0, 
			simpleHasher->wndMain, NULL, hInstance, NULL);	
	if ( ! simpleHasher->lblSelectDrive)
		return 3;
		
	/*Start button */
	simpleHasher->btnStart = CreateWindowEx(0, "BUTTON", "Inizia", WS_CHILD | WS_VISIBLE | BS_FLAT | BS_PUSHBUTTON, 0, 0, 0, 0, simpleHasher->wndMain, NULL, hInstance, NULL);	
	if ( ! simpleHasher->btnStart)
		return 4;	
		
	/* Checkboxes */
	simpleHasher->chkShowPercent = CreateWindowEx(0, "BUTTON", "Calcola percentuale (pi lento)", WS_CHILD | WS_VISIBLE | BS_FLAT | BS_AUTOCHECKBOX, 0, 0, 0, 0, simpleHasher->wndMain, NULL, hInstance, NULL);	
	if ( ! simpleHasher->chkShowPercent)
		return 5;
		
	simpleHasher->chkEncrypted = CreateWindowEx(0, "BUTTON", "Le evidenze sono cifrate", WS_CHILD /*| WS_VISIBLE*/ | BS_FLAT | BS_AUTOCHECKBOX, 0, 0, 0, 0, simpleHasher->wndMain, NULL, hInstance, NULL);	
	if ( ! simpleHasher->chkEncrypted)
		return 6;	
		
	/* File browse */
	simpleHasher->txtFilePath = CreateWindowEx(0, "EDIT", "", WS_CHILD | WS_VISIBLE | ES_AUTOHSCROLL | ES_LEFT | WS_BORDER, 0, 0, 0, 0, simpleHasher->wndMain, NULL, hInstance, NULL);	
	if ( ! simpleHasher->txtFilePath)
		return 7;
		
	simpleHasher->btnBrowse = CreateWindowEx(0, "BUTTON", "Sfoglia", WS_CHILD | WS_VISIBLE | BS_FLAT | BS_PUSHBUTTON, 0, 0, 0, 0, simpleHasher->wndMain, NULL, hInstance, NULL);	
	if ( ! simpleHasher->btnBrowse)
		return 8;	
		
	/* Drive list*/
	simpleHasher->tvwDrives = CreateWindowEx(0, WC_TREEVIEW, "", WS_VISIBLE | WS_CHILD | WS_BORDER | TVS_CHECKBOXES | TVS_DISABLEDRAGDROP | TVS_NOHSCROLL,  0, 0, 0,0, simpleHasher->wndMain, NULL, hInstance, NULL);
	if ( ! simpleHasher->tvwDrives)
		return 9;	
		
	/* Label*/
	simpleHasher->lblSelectFile = CreateWindowEx(0, "STATIC", "Seleziona il file delle evidenze o trascinalo qui sopra", WS_CHILD | WS_VISIBLE,  0, 0, 0,0, simpleHasher->wndMain, NULL, hInstance, NULL);
	if ( ! simpleHasher->lblSelectFile)
		return 10;	
		
	/* Result*/
	simpleHasher->txtLog = CreateWindowEx(0, "EDIT", "", WS_CHILD | WS_VSCROLL | WS_HSCROLL  | ES_LEFT | ES_MULTILINE  | ES_READONLY , 
		 0, 0, 0,0, simpleHasher->wndMain, NULL, hInstance, NULL);
	if ( ! simpleHasher->txtLog)
		return 11;		
		
	/* Pause */
	simpleHasher->btnPause = CreateWindowEx(0, "BUTTON", "Sospendi", WS_CHILD  | BS_FLAT | BS_PUSHBUTTON, 0, 0, 0, 0, simpleHasher->wndMain, NULL, hInstance, NULL);	
	if ( ! simpleHasher->btnPause)
		return 12;	
		
	/* Label progress*/
	simpleHasher->lblProgress = CreateWindowEx(0, "STATIC", "", WS_CHILD,  0, 0, 0,0, simpleHasher->wndMain, NULL, hInstance, NULL);
	if ( ! simpleHasher->lblProgress)
		return 13;		
		
	/* Set the fonts */	
	SendMessage(simpleHasher->lblSelectDrive, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->btnStart, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->chkShowPercent, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->chkEncrypted, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->txtFilePath, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->btnBrowse, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->tvwDrives, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->lblSelectFile, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->btnPause, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->txtLog, WM_SETFONT, (WPARAM)hFont, 0);
	SendMessage(simpleHasher->lblProgress, WM_SETFONT, (WPARAM)hFont, 0);
	
	/* Show the main window*/	
	ShowWindow(simpleHasher->wndMain, SW_SHOW);
	
	return 0;
}

/* Windows' message loop */
void MessageLoop(HINSTANCE hInstance)
{
	(void)hInstance;
	
	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

/* Dispatch the messages */
LRESULT WINAPI TableWinProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	struct MSG_HANDLER* table = (struct MSG_HANDLER*)GetWindowLongPtr(hWnd, 0);
	
		for (; table && table->proc; table++)
		if (Msg == table->msg)
			return table->proc(hWnd, Msg, wParam, lParam);
		
	return DefWindowProc(hWnd, Msg, wParam, lParam);
}
