#include <windows.h>
#include "guiutil.h"

/* Variable length */
struct FINFO
{
	wchar_t filename[MAX_FILENAME + 1];
	HANDLE hFile;
};

struct FQUEUE 
{
	volatile unsigned int capacity;
	volatile unsigned int nextFree;
	volatile unsigned int firstUsed;
	CRITICAL_SECTION cs;
	volatile unsigned int n;
	volatile BOOL stopped;
	struct FINFO handles[256];
};


void InitFQueue(struct FQUEUE** q, int capacity)
{
	*q = AllocMem(sizeof(struct FQUEUE));
	if ( ! *q)
		return;
	
	struct FQUEUE* p = *q;
		
	p->capacity = capacity;
	p->nextFree = p->firstUsed = p->n = p->stopped = 0;
	InitializeCriticalSection(&p->cs);
}

void FreeFQueue(struct FQUEUE* q)
{
	DeleteCriticalSection(&q->cs);
	FreeMem(q);
}

BOOL AddItem(struct FQUEUE* q, HANDLE handle, wchar_t* filename)
{
	EnterCriticalSection(&q->cs);
	BOOL ret = TRUE;
	
	/* Ignore new adding if stopped */
	if ( q->stopped)
		goto leave;
		
	/* Full? */	
	if (q->n >= q->capacity)
	{
		ret = FALSE;
		goto leave;
	}
		
	/* Add */
	q->handles[q->nextFree].hFile = handle;
	wcscpy(q->handles[q->nextFree].filename, filename);
	q->nextFree = (q->nextFree + 1) % q->capacity;
	q->n++;
	
leave:
	LeaveCriticalSection(&q->cs);
	
	return ret;
}

BOOL GetItem(struct FQUEUE* q, wchar_t* filename, HANDLE* pHandle)
{
	BOOL ret = TRUE;
	EnterCriticalSection(&q->cs);
	
	if (q->n == 0 && q->stopped)
	{
		*pHandle = INVALID_HANDLE_VALUE;
		wcscpy(filename, L"\0");
	}
	else
	{
		if (q->n == 0)
			ret = FALSE;
		else
		{
			*pHandle = q->handles[q->firstUsed].hFile;
			wcscpy(filename, q->handles[q->firstUsed].filename);
			q->firstUsed = (q->firstUsed + 1) % q->capacity;
			q->n--;
		}
	}
	
	LeaveCriticalSection(&q->cs);
	
	return ret;		
}

void ResetQueue(struct FQUEUE* q)
{
	EnterCriticalSection(&q->cs);
	q->nextFree = q->firstUsed = q->n = q->stopped = 0;
	LeaveCriticalSection(&q->cs);
}

void StopQueue(struct FQUEUE* q, BOOL flush)
{
	EnterCriticalSection(&q->cs);
	if (flush)
		ResetQueue(q);
	q->stopped = TRUE;	
	LeaveCriticalSection(&q->cs);
}
